/*
        The original file was written 1996-1999 by Donald Becker.

	This software may be used and distributed according to the terms
	of the GNU General Public License, incorporated herein by reference.

	This driver is for the 3Com "Vortex" and "Boomerang" series ethercards.
	Members of the series include Fast EtherLink 3c590/3c592/3c595/3c597
	and the EtherLink XL 3c900 and 3c905 cards.

	Problem reports and questions should be directed to
	vortex@scyld.com

	The author may be reached as becker@scyld.com, or C/O
	Scyld Computing Corporation
	410 Severn Ave., Suite 210
	Annapolis MD 21403

	Linux Kernel Additions (better to see the last one first):
	
 	0.99H+lk0.9 - David S. Miller - softnet, PCI DMA updates
 	0.99H+lk1.0 - Jeff Garzik <jgarzik@mandrakesoft.com>
		Remove compatibility defines for kernel versions < 2.2.x.
		Update for new 2.3.x module interface
	LK1.1.2 (March 19, 2000)
	* New PCI interface (jgarzik)

    LK1.1.3 25 April 2000, Andrew Morton <andrewm@uow.edu.au>
    - Merged with 3c575_cb.c
    - Don't set RxComplete in boomerang interrupt enable reg
    - spinlock in vortex_timer to protect mdio functions
    - disable local interrupts around call to vortex_interrupt in
      vortex_tx_timeout() (So vortex_interrupt can use spin_lock())
    - Select window 3 in vortex_timer()'s write to Wn3_MAC_Ctrl
    - In vortex_start_xmit(), move the lock to _after_ we've altered
      vp->cur_tx and vp->tx_full.  This defeats the race between
      vortex_start_xmit() and vortex_interrupt which was identified
      by Bogdan Costescu.
    - Merged back support for six new cards from various sources
    - Set vortex_have_pci if pci_module_init returns zero (fixes cardbus
      insertion oops)
    - Tell it that 3c905C has NWAY for 100bT autoneg
    - Fix handling of SetStatusEnd in 'Too much work..' code, as
      per 2.3.99's 3c575_cb (Dave Hinds).
    - Split ISR into two for vortex & boomerang
    - Fix MOD_INC/DEC races
    - Handle resource allocation failures.
    - Fix 3CCFE575CT LED polarity
    - Make tx_interrupt_mitigation the default

    LK1.1.4 25 April 2000, Andrew Morton <andrewm@uow.edu.au>    
    - Add extra TxReset to vortex_up() to fix 575_cb hotplug initialisation probs.
    - Put vortex_info_tbl into __devinitdata
    - In the vortex_error StatsFull HACK, disable stats in vp->intr_enable as well
      as in the hardware.
    - Increased the loop counter in issue_and_wait from 2,000 to 4,000.

    LK1.1.5 28 April 2000, andrewm
    - Added powerpc defines (John Daniel <jdaniel@etresoft.com> said these work...)
    - Some extra diagnostics
    - In vortex_error(), reset the Tx on maxCollisions.  Otherwise most
      chips usually get a Tx timeout.
    - Added extra_reset module parm
    - Replaced some inline timer manip with mod_timer
      (Franois romieu <Francois.Romieu@nic.fr>)
    - In vortex_up(), don't make Wn3_config initialisation dependent upon has_nway
      (this came across from 3c575_cb).

    LK1.1.6 06 Jun 2000, andrewm
    - Backed out the PPC defines.
    - Use del_timer_sync(), mod_timer().
    - Fix wrapped ulong comparison in boomerang_rx()
    - Add IS_TORNADO, use it to suppress 3c905C checksum error msg
      (Donald Becker, I Lee Hetherington <ilh@sls.lcs.mit.edu>)
    - Replace union wn3_config with BFINS/BFEXT manipulation for
      sparc64 (Pete Zaitcev, Peter Jones)
    - In vortex_error, do_tx_reset and vortex_tx_timeout(Vortex):
      do a netif_wake_queue() to better recover from errors. (Anders Pedersen,
      Donald Becker)
    - Print a warning on out-of-memory (rate limited to 1 per 10 secs)
    - Added two more Cardbus 575 NICs: 5b57 and 6564 (Paul Wagland)

    LK1.1.7 2 Jul 2000 andrewm
    - Better handling of shared IRQs
    - Reset the transmitter on a Tx reclaim error
    - Fixed crash under OOM during vortex_open() (Mark Hemment)
    - Fix Rx cessation problem during OOM (help from Mark Hemment)
    - The spinlocks around the mdio access were blocking interrupts for 300uS.
      Fix all this to use spin_lock_bh() within mdio_read/write
    - Only write to TxFreeThreshold if it's a boomerang - other NICs don't
      have one.
    - Added 802.3x MAC-layer flow control support

   LK1.1.8 13 Aug 2000 andrewm
    - Ignore request_region() return value - already reserved if Cardbus.
    - Merged some additional Cardbus flags from Don's 0.99Qk
    - Some fixes for 3c556 (Fred Maciel)
    - Fix for EISA initialisation (Jan Rekorajski)
    - Renamed MII_XCVR_PWR and EEPROM_230 to align with 3c575_cb and D. Becker's drivers
    - Fixed MII_XCVR_PWR for 3CCFE575CT
    - Added INVERT_LED_PWR, used it.
    - Backed out the extra_reset stuff

   LK1.1.9 12 Sep 2000 andrewm
    - Backed out the tx_reset_resume flags.  It was a no-op.
    - In vortex_error, don't reset the Tx on txReclaim errors
    - In vortex_error, don't reset the Tx on maxCollisions errors.
      Hence backed out all the DownListPtr logic here.
    - In vortex_error, give Tornado cards a partial TxReset on
      maxCollisions (David Hinds).  Defined MAX_COLLISION_RESET for this.
    - Redid some driver flags and device names based on pcmcia_cs-3.1.20.
    - Fixed a bug where, if vp->tx_full is set when the interface
      is downed, it remains set when the interface is upped.  Bad
      things happen.

   LK1.1.10 17 Sep 2000 andrewm
    - Added EEPROM_8BIT for 3c555 (Fred Maciel)
    - Added experimental support for the 3c556B Laptop Hurricane (Louis Gerbarg)
    - Add HAS_NWAY to "3c900 Cyclone 10Mbps TPO"

   LK1.1.11 13 Nov 2000 andrewm
    - Dump MOD_INC/DEC_USE_COUNT, use SET_MODULE_OWNER

   LK1.1.12 1 Jan 2001 andrewm (2.4.0-pre1)
    - Call pci_enable_device before we request our IRQ (Tobias Ringstrom)
    - Add 3c590 PCI latency timer hack to vortex_probe1 (from 0.99Ra)
    - Added extended issue_and_wait for the 3c905CX.
    - Look for an MII on PHY index 24 first (3c905CX oddity).
    - Add HAS_NWAY to 3cSOHO100-TX (Brett Frankenberger)
    - Don't free skbs we don't own on oom path in vortex_open().

   LK1.1.13 27 Jan 2001
    - Added explicit `medialock' flag so we can truly
      lock the media type down with `options'.
    - "check ioremap return and some tidbits" (Arnaldo Carvalho de Melo <acme@conectiva.com.br>)
    - Added and used EEPROM_NORESET for 3c556B PM resumes.
    - Fixed leakage of vp->rx_ring.
    - Break out separate HAS_HWCKSM device capability flag.
    - Kill vp->tx_full (ANK)
    - Merge zerocopy fragment handling (ANK?)

   LK1.1.14 15 Feb 2001
    - Enable WOL.  Can be turned on with `enable_wol' module option.
    - EISA and PCI initialisation fixes (jgarzik, Manfred Spraul)
    - If a device's internalconfig register reports it has NWAY,
      use it, even if autoselect is enabled.

   LK1.1.15 6 June 2001 akpm
    - Prevent double counting of received bytes (Lars Christensen)
    - Add ethtool support (jgarzik)
    - Add module parm descriptions (Andrzej M. Krzysztofowicz)
    - Implemented alloc_etherdev() API
    - Special-case the 'Tx error 82' message.

   LK1.1.16 18 July 2001 akpm
    - Make NETIF_F_SG dependent upon nr_free_highpages(), not on CONFIG_HIGHMEM
    - Lessen verbosity of bootup messages
    - Fix WOL - use new PM API functions.
    - Use netif_running() instead of vp->open in suspend/resume.
    - Don't reset the interface logic on open/close/rmmod.  It upsets
      autonegotiation, and hence DHCP (from 0.99T).
    - Back out EEPROM_NORESET flag because of the above (we do it for all
      NICs).
    - Correct 3c982 identification string
    - Rename wait_for_completion() to issue_and_wait() to avoid completion.h
      clash.


=====================================================================================
                          PORTING TO RT-LINUX                       
=====================================================================================

   LK1.1.17 14 March 2003 

            Sergio Perez Alcaiz  <serpeal@upvnet.upv.es>  Departamento
            de  Informtica  de  Sistemas  y Computadores  Universidad
            Politcnica de Valencia

	    The RTL-lwIP project  has  been supported by  the  Spanish
	    Government  Research   Office (CICYT)     under      grant
	    TIC2002-04123-C03-03   SISTEMAS DE TIEMPO REAL EMPOTRADOS,
	    FIABLES Y DISTRIBUIDOS BASADOS EN COMPONENTES

          Valencia (Spain)                      

	    Porting the driver to RT-Linux, that is:
	    -Giving to  the  driver the RT-Linux  driver  architecture
	    (open, close, read, write, ioctl).
	    -Using my own buffer management.
	    -No NIC stats allowed.
	    -Some calls of pci stuff have been implemented.
	    -Changing  the    interrupts management and    the handler
	    phylosophy.
	    -Changing the "receive" and "send" phylosophy.
	    -Introducing signaling.
	    -Registering threads  that  want  to  be notified  of  the
	    arrival of a packet.
	    -IP filtering.
	    -Ethernet filtering.
	     - ....

	     The porting is just for  the 3Com905C-X card. So no other
	     cards are supported (as far as I know...).
	     
	     Just take a look at it.

   LK1.1.18     12     January    2004

            Sergio Perez Alcaiz <serpeal@upvnet.upv.es> Departamento
            de Informtica de Sistemas y Computadores Universidad
            Politcnica de Valencia

	    The  REDD    project has  been   supported by  the Spanish
	    Government    Research  Office    (CICYT)  under     grant
	    TIC2002-04123-C03-03   SISTEMAS DE TIEMPO REAL EMPOTRADOS,
	    FIABLES  Y DISTRIBUIDOS  BASADOS  EN COMPONENTES  Valencia
	    (Spain)

	    -Removed threads  registration (to  better comply with the
	    POSIX standard).
            -Added  a semaphore   to block   threads  when they    are
            performing a read and there are not packets available.
	    -Modified to use the  new release of DIDMA (Dynamic Memory
	    Allocator) now called TLSF.

   LK1.1.19     1     December    2004

            Sergio Perez Alcaiz <serpeal@upvnet.upv.es> Departamento
            de Informtica de Sistemas y Computadores Universidad
            Politcnica de Valencia

	    - Driver  modified  to provide the   O_NOBLOCK access mode
	    (non-blocking access).  
	    -  Driver modified to  support  the zero copy technique or
	    not, depending on the options set in the reddopts.h.
	    -   Driver  modified to  support   the   FAKE architecture
	    described in the reddoopts.h

=====================================================================================

    - see Documentation/networking/vortex.txt
*/


MODULE_LICENSE("GPL");

static int rt_3c905c_interrupted = 0, rt_3c905c_writting = 0;


/***************************************************************************************/
/* This function is called in the cleanup_module function, and it must do other stuff  */
/* not done in the cleanup_module function in order to "close" the driver. In some     */
/* implementations this function could be empty.                                       */
/***************************************************************************************/
static void other_stuff_closing(void){
  rtl_3COM905C_issue_and_wait(UpStall);
  rtl_3COM905C_issue_and_wait(DownStall);
}

/***************************************************************************************/
/* This is the interrupt handler. It is executed when the ethernet card raises an      */
/* interrupt and interrupts are enabled. In the initialisation phase we've forced the  */
/* card to only generate the UpComplete interrupt, that is, when a full incoming       */
/* packet has been processed.                                                          */
/***************************************************************************************/
unsigned int boomerang_interrupt(unsigned int irq , struct pt_regs *regs){
  long ioaddr = private_data.ioaddr;
  int status = inw(ioaddr + EL3_STATUS);
  int work_done = max_interrupt_work;
  struct pci_dev *dev = eth_dev;

  inside_the_interrupt_handler = 1;
  rt_3c905c_interrupted++;
  if(rt_3c905c_writting & rt_3c905c_interrupted)
    rtl_printf("I've been interrupted %d times\n",rt_3c905c_interrupted);

  if (status & UpComplete) {
    private_data.rx_packets++;    
    rtl_3COM905C_issue_and_wait(UpStall);
    while(private_data.rx_ring[private_data.cur_rx % RX_RING_SIZE].status & 0x00008000){
      boomerang_rx(dev);
      outw(AckIntr | UpComplete, ioaddr + EL3_CMD);
      if (--work_done < 0)
	break;
    }
    outw(UpUnstall, ioaddr + EL3_CMD);
  }
  
  if (status & DownComplete){
    rtl_printf("Se acaba de enviar el paquete\n");
    outw(AckIntr | DownComplete, ioaddr + EL3_CMD);
  }
  
  /* Check for all uncommon interrupts at once. */
  if (status & (HostError | RxEarly | StatsFull | IntReq)){
    rtl_printf("VORTEX_ERROR\n");
    vortex_error(dev, status);
  }
  
  /* Acknowledge the IRQ. */
  outw(AckIntr | IntReq | IntLatch, ioaddr + EL3_CMD);
  if (private_data.cb_fn_base){	
    writel(0x8000, private_data.cb_fn_base + 4);
  }
  
  /* We must be sure that we're out of the interrupt handler before cleanup_modules */
  /* is executed. If cleanup_modules is being executed, we don't have to enable the */
  /* irq. If enabled, then the system could crash.                                  */
  if(!trying_to_close)
    rtl_hard_enable_irq(private_data.pdev->irq);  
  
  rt_3c905c_interrupted--;

  return (inside_the_interrupt_handler = 0);
}

/***************************************************************************************/
/* This is the function called when an incoming packet has succesfuly arrived, so we   */
/* have to copy it into internal buffers.                                              */
/***************************************************************************************/
static int boomerang_rx(struct pci_dev *dev)
{
  int entry = private_data.cur_rx % RX_RING_SIZE;
  int rx_status;
  int rx_work_limit = private_data.dirty_rx + RX_RING_SIZE - private_data.cur_rx;
  unsigned char *buffer;
  int temp = private_data.rx_ring[entry].status & 0x1fff;

  buffer =private_data.rx_skbuff[entry];

  filter_incoming_packet(buffer, temp);

  while ((rx_status = le32_to_cpu(private_data.rx_ring[entry].status)) & RxDComplete){

    if (--rx_work_limit < 0)
      break;
    if (!(rx_status & RxDError))  /* Error, update stats. */
      private_data.stats.rx_packets++;
  }

  entry = (++private_data.cur_rx) % RX_RING_SIZE;

  /* Refill the Rx ring buffers. */
  for (; private_data.cur_rx - private_data.dirty_rx > 0; private_data.dirty_rx++) {
    entry = private_data.dirty_rx % RX_RING_SIZE;
    private_data.rx_ring[entry].status = 0;	/* Clear complete bit. */
  }
  return 0;
}

/***************************************************************************************/
/* This function does internal initialisations of the card. It writes into internal    */
/* registers and checks the card capabilities.                                         */
/***************************************************************************************/
static int start_up_device(struct pci_dev *dev){
  struct vortex_chip_info * const vci = &vortex_info;
  unsigned int eeprom[0x40], checksum = 0;		/* EEPROM contents */
  char *print_name;
  int retval;
  long ioaddr;
  int i,step;

  print_name = dev ? dev->slot_name : "3c59x";

  ioaddr = rt_pci_resource_start(dev, 0);

  private_data.drv_flags = vci->drv_flags;
  private_data.has_nway = (vci->drv_flags & HAS_NWAY) ? 1 : 0;
  private_data.io_size = vci->io_size;
  private_data.card_idx = 0;
  private_data.ioaddr = ioaddr;
  private_data.media_override = 7;
  private_data.mtu = mtu;

  print_name = dev ? dev->slot_name : "3c59x";

  /* PCI-only startup logic */
  if (dev) {
    if (request_region(ioaddr, vci->io_size, print_name) != NULL){
      private_data.must_free_region = 1;
     }

    /* enable bus-mastering if necessary */		
    if (vci->flags & PCI_USES_MASTER)
      pci_set_master (dev);

    private_data.pdev = dev;

    /* Makes sure rings are at least 16 byte aligned. */
    private_data.rx_ring = pci_alloc_consistent(dev, sizeof(struct boom_rx_desc) * RX_RING_SIZE, &private_data.rx_ring_dma);
    private_data.tx_ring = pci_alloc_consistent(dev, sizeof(struct boom_tx_desc) * TX_RING_SIZE, &private_data.tx_ring_dma);

    retval = -ENOMEM;
    if ((private_data.rx_ring == 0) || (private_data.tx_ring == 0))
      goto free_region;
        
    EL3WINDOW(0);
    {
      int base;

      if (vci->drv_flags & EEPROM_8BIT)
	base = 0x230;
      else if (vci->drv_flags & EEPROM_OFFSET)
	base = EEPROM_Read + 0x30;
      else
	base = EEPROM_Read;

      for (i = 0; i < 0x40; i++) {
	int timer;
	
	/* This means that we want to read EepromCommand Register and disable writting */
	/* Issuing ReadRegister & WriteDisable                                         */
	outw(base + i, ioaddr + Wn0EepromCmd);

	for (timer = 10; timer >= 0; timer--) {

	  /* The read data is available through the EepromData register 162us after  */
	  /* the ReadRegister command has been issued                                */
	  rtl_delay(162000);

	  /* Bit 15th (eepromBusy) of EepromCommand register is a read-only bit asserted   */
	  /* during the execution of EEProm commands. Further commans should not be issued */
	  /* to the EepromCommand register, nor should data be read from the EepromData    */
	  /* register while this bit is true                                               */
	  if ((inw(ioaddr + Wn0EepromCmd) & 0x8000) == 0)
	    break;
	}

	/* Save the contents of the 3C90xC NIC's EEPROM     */ 
	eeprom[i] = inw(ioaddr + Wn0EepromData);
      }

    }//EL3WINDOW(0) configuration finished

    /* EEPROM can be checksummed in order to assure that reading was OK */
    for (i = 0; i < 0x18; i++)
      checksum ^= eeprom[i];
    checksum = (checksum ^ (checksum >> 8)) & 0xff;
    if (checksum != 0x00) {		/* Grrr, needless incompatible change 3Com. */
      while (i < 0x21)
	checksum ^= eeprom[i++];
      checksum = (checksum ^ (checksum >> 8)) & 0xff;
    }
    if ((checksum != 0x00) && !(vci->drv_flags & IS_TORNADO))
      printk(" ***INVALID CHECKSUM %4.4x*** ", checksum);
    
    /* Save HW address into dev_addr (MAC address in format 00:04:75:bd:ea:e7) */
    for (i = 0; i < 3; i++)
      ((u16 *)private_data.dev_addr)[i] = htons(eeprom[i + 10]);
    
    /* This writes into the StationAddress register the NIC's HW address in order */
    /* to define the individual destination address that the NIC responds to when */
    /* receiving packets                                                          */
    EL3WINDOW(2);
    for (i = 0; i < 6; i++)
      outb(private_data.dev_addr[i], ioaddr + i);

    EL3WINDOW(4);
    step = (inb(ioaddr + Wn4_NetDiag) & 0x1e) >> 1;

    if (dev && vci->drv_flags & HAS_CB_FNS) {
      unsigned long fn_st_addr;			/* Cardbus function status space */
      unsigned short n;
      
      fn_st_addr = pci_resource_start (dev, 2);
      if (fn_st_addr) {
	private_data.cb_fn_base = ioremap(fn_st_addr, 128);
	retval = -ENOMEM;
	if (!private_data.cb_fn_base)
	  goto free_ring;
      }

      EL3WINDOW(2);
      
      n = inw(ioaddr + Wn2_ResetOptions) & ~0x4010;
      if (private_data.drv_flags & INVERT_LED_PWR)
	n |= 0x10;
      if (private_data.drv_flags & INVERT_MII_PWR)
	n |= 0x4000;
      outw(n, ioaddr + Wn2_ResetOptions);
    }

    /* Extract our information from the EEPROM data. */
    private_data.info1 = eeprom[13];
    private_data.info2 = eeprom[15];
    private_data.capabilities = eeprom[16];
 
    if (private_data.info1 & 0x8000){
      private_data.full_duplex = 1;
      printk(KERN_INFO "Full duplex capable\n");
    }

    {
      unsigned int config;
      EL3WINDOW(3);

      /* This reads the MediaOptions register which shows what physical media */
      /* connections are available in the NIC                                 */
      private_data.available_media = inw(ioaddr + Wn3_Options); //Wn3_Options = 8 vp.available_media = 0xa

      if ((private_data.available_media & 0xff) == 0)		/* Broken 3c916 */
	private_data.available_media = 0x40;

      /* This reads the InternalConfig register which provides a way to set */
      /* NIC-specific, non-host-related configuration settings              */ 
      config = inl(ioaddr + Wn3_Config); //Wn3_Config = 0

      private_data.default_media = XCVR(config);
      if (private_data.default_media == XCVR_NWAY)
	private_data.has_nway = 1;
      private_data.autoselect = AUTOSELECT(config);
    }

    if (private_data.media_override != 7) {
      private_data.if_port = private_data.media_override;
    } else
      private_data.if_port = private_data.default_media;

    if (private_data.if_port == XCVR_MII || private_data.if_port == XCVR_NWAY) {
      int phy, phy_idx = 0;
      EL3WINDOW(4);
      mii_preamble_required++;
      mii_preamble_required++;
      rtl_3COM905C_mdio_read(24, 1);
      for (phy = 0; phy < 32 && phy_idx < 1; phy++) {
	int mii_status, phyx;
	
	/*
	 * For the 3c905CX we look at index 24 first, because it bogusly
	 * reports an external PHY at all indices
	 */
	if (phy == 0)
	  phyx = 24;
	else if (phy <= 24)
	  phyx = phy - 1;
	else
	  phyx = phy;
	mii_status = rtl_3COM905C_mdio_read(phyx, 1);
	if (mii_status  &&  mii_status != 0xffff) {
	  private_data.phys[phy_idx++] = phyx;
	  
	  if ((mii_status & 0x0040) == 0)
	    mii_preamble_required++;
	}
      }		
      mii_preamble_required--;
      if (phy_idx == 0) {
	private_data.phys[0] = 24;
      } else {
	private_data.advertising = rtl_3COM905C_mdio_read(private_data.phys[0], 4);
	if (private_data.full_duplex) {
	  /* Only advertise the FD media types. */
	  private_data.advertising &= ~0x02A0;
	  rtl_3COM905C_mdio_write(private_data.phys[0], 4, private_data.advertising);
	}
      }
    }
    if (private_data.capabilities & CapBusMaster) {
      private_data.full_bus_master_tx = 1;
      
      private_data.full_bus_master_rx = (private_data.info2 & 1) ? 1 : 2;
      private_data.bus_master = 0;		/* AKPM: vortex only */
    }
    if (private_data.pdev && private_data.enable_wol) {
      private_data.pm_state_valid = 1;
      rt_pci_save_state(dev, private_data.power_state);
      rtl_3COM905C_acpi_set_WOL();
    }
    vortex_open(dev);

  }// if(dev)
    
  return 0;

 free_ring:

  pci_free_consistent(dev, sizeof(struct boom_rx_desc) * RX_RING_SIZE, private_data.rx_ring, private_data.rx_ring_dma);
  pci_free_consistent(dev, sizeof(struct boom_tx_desc) * TX_RING_SIZE, private_data.tx_ring, private_data.tx_ring_dma);

  return -1;

 free_region:
  if (private_data.must_free_region){
    release_region(ioaddr, vci->io_size);
    private_data.must_free_region = 0;
  }
  printk("vortex_probe1 fails.  Returns %d\n", retval);

  return -1;
}

/***************************************************************************************/
/* ACPI: Advanced Configuration and Power Interface.                                   */
/* Set Wake-On-LAN mode and put the board into D3 (power-down) state.                  */
/***************************************************************************************/
static void rtl_3COM905C_acpi_set_WOL(void)
{
  long ioaddr = private_data.ioaddr;

  /* Power up on: 1==Downloaded Filter, 2==Magic Packets, 4==Link Status. */
  EL3WINDOW(7);
  outw(2, ioaddr + 0x0c);
  /* The RxFilter must accept the WOL frames. */
  outw(SetRxFilter|RxStation|RxBroadcast, ioaddr + EL3_CMD); //RxMulticast
  outw(RxEnable, ioaddr + EL3_CMD);

  /* Change the power state to D3; RxEnable doesn't take effect. */
  rt_pci_enable_wake(private_data.pdev, 0, 1);
  rt_pci_set_power_state(private_data.pdev, 3);
}

/***************************************************************************************/
/* This function registers the interrupt handler and reserves memory for buffers which */
/* only will be accessed by the card.                                                  */
/***************************************************************************************/
static int vortex_open(struct pci_dev *dev)
{
  int i;
  int retval;
  
  /* Use the now-standard shared IRQ implementation. */
  if ((retval = rtl_request_irq(dev->irq, &boomerang_interrupt))) {
    printk(KERN_ERR "%s: Could not reserve IRQ %d\n", dev->name, dev->irq);
    goto out;
  }else
    private_data.must_free_irq = 1;

  if (private_data.full_bus_master_rx) { /* Boomerang bus master. */

    /* RX RING INITIALIZATION */
    for (i = 0; i < RX_RING_SIZE; i++) {
      unsigned char *skb;
      private_data.rx_ring[i].next = cpu_to_le32(private_data.rx_ring_dma + sizeof(struct boom_rx_desc) * (i+1));
      private_data.rx_ring[i].status = 0;	/* Clear complete bit. */
      private_data.rx_ring[i].length = cpu_to_le32(PKT_BUF_SZ | LAST_FRAG);
      skb = rtl_malloc(PKT_BUF_SZ);
      private_data.rx_skbuff[i] = skb;
      if (skb == NULL)
	break;			/* Bad news!  */
      private_data.rx_ring[i].addr = cpu_to_le32(pci_map_single(private_data.pdev, skb, PKT_BUF_SZ, PCI_DMA_FROMDEVICE));
    }

    if (i != RX_RING_SIZE) {
      int j;
      printk(KERN_EMERG "%s: no memory for rx ring\n", dev->name);
      for (j = 0; j < i; j++) {
	if (private_data.rx_skbuff[j]) {
	  rtl_free(private_data.rx_skbuff[j]);
	  private_data.rx_skbuff[j] = 0;
	}
      }
      retval = -ENOMEM;
      goto out_free_irq;
    }

    /* Wrap the ring. */
    private_data.rx_ring[i-1].next = cpu_to_le32(private_data.rx_ring_dma);

    /* TX RING INITIALIZATION */
    for (i = 0; i < TX_RING_SIZE; i++) {
      unsigned char *skb;
      private_data.tx_ring[i].next = 0; 
      private_data.tx_ring[i].status = 0;	/* Clear complete bit. */
      private_data.tx_ring[i].length = cpu_to_le32(PKT_BUF_SZ | LAST_FRAG);
      skb = rtl_malloc(PKT_BUF_SZ);
      private_data.tx_skbuff[i] = skb;
      if (skb == NULL)
	break;			/* Bad news!  */
      private_data.tx_ring[i].addr = cpu_to_le32(pci_map_single(private_data.pdev, skb, PKT_BUF_SZ, PCI_DMA_FROMDEVICE));
    }

    
    if (i != TX_RING_SIZE) {
      int j;
      printk(KERN_EMERG "%s: no memory for tx ring\n", dev->name);
      for (j = 0; j < i; j++) {
	if (private_data.tx_skbuff[j]) {
	  rtl_free(private_data.tx_skbuff[j]);
	  private_data.tx_skbuff[j] = 0;
	}
      }
      retval = -ENOMEM;
      goto out_free_irq;
    }

    /* Wrap the ring. */
    private_data.tx_ring[i-1].next = cpu_to_le32(private_data.tx_ring_dma);
  }

  vortex_up(dev);
  return 0;
  
 out_free_irq:
  rtl_free_irq(dev->irq); 
 out:
  return retval;
}

/***************************************************************************************/
/* This function is also used to initialise the card. In this function we indicate     */
/* which interrupts do we want the card to generate.                                   */
/***************************************************************************************/
static void vortex_up(struct pci_dev *dev)
{
  long ioaddr = private_data.ioaddr;
  unsigned int config;
  int i;
  
  private_data.rx_packets = 0;
  private_data.rx_frames_for_us = 0;

  if (private_data.pdev && private_data.enable_wol) {
    rt_pci_set_power_state(dev, 0);	/* Go active */
    rt_pci_restore_state(dev, private_data.power_state);
  }
  
  /* Before initializing select the active media port. */
  EL3WINDOW(3);
  config = inl(ioaddr + Wn3_Config);

  if (private_data.media_override != 7) {
    private_data.if_port = private_data.media_override;
  } else if (private_data.autoselect) {
    if (private_data.has_nway) {
    } else {
      /* Find first available media type, starting with 100baseTx. */
      private_data.if_port = XCVR_100baseTx;
      while (! (private_data.available_media & media_tbl[private_data.if_port].mask))
	private_data.if_port = media_tbl[private_data.if_port].next;
    }
  } else {
    private_data.if_port = private_data.default_media;
    
  }

  private_data.full_duplex = private_data.force_fd;

  config = BFINS(config, private_data.if_port, 20, 4);

  outl(config, ioaddr + Wn3_Config);
  
  if (private_data.if_port == XCVR_MII || private_data.if_port == XCVR_NWAY) {
    int mii_reg1, mii_reg5;
    EL3WINDOW(4);
    /* Read BMSR (reg1) only to clear old status. */
    mii_reg1 = rtl_3COM905C_mdio_read(private_data.phys[0], 1);
    mii_reg5 = rtl_3COM905C_mdio_read(private_data.phys[0], 5);
    if (mii_reg5 == 0xffff  ||  mii_reg5 == 0x0000)
      ;					/* No MII device or no link partner report */
    else if ((mii_reg5 & 0x0100) != 0	/* 100baseTx-FD */
	     || (mii_reg5 & 0x00C0) == 0x0040) /* 10T-FD, but not 100-HD */
      private_data.full_duplex = 1;
    private_data.partner_flow_ctrl = ((mii_reg5 & 0x0400) != 0);

    EL3WINDOW(3);
  }

  private_data.full_duplex = 1;

  /* Set the full-duplex bit. */

  outw(	((private_data.info1 & 0x8000) || private_data.full_duplex ? 0x20 : 0) |
	(private_data.mtu > 1500 ? 0x40 : 0) |
	((private_data.full_duplex && private_data.flow_ctrl && private_data.partner_flow_ctrl) ? 0x100 : 0),
	ioaddr + Wn3_MAC_Ctrl);
  
  rtl_3COM905C_issue_and_wait(TxReset);

  /*
   * Don't reset the PHY - that upsets autonegotiation during DHCP operations.
   */
  rtl_3COM905C_issue_and_wait(RxReset|0x04);
  
  outw(SetStatusEnb | 0x00, ioaddr + EL3_CMD);

  EL3WINDOW(4);

  /* Set the station address and mask in window 2 each time opened. */
  EL3WINDOW(2);
  for (i = 0; i < 6; i++)
    outb(private_data.dev_addr[i], ioaddr + i);
  for (; i < 12; i+=2)
    outw(0, ioaddr + i);
  
  if (private_data.cb_fn_base) {
    unsigned short n = inw(ioaddr + Wn2_ResetOptions) & ~0x4010;
    if (private_data.drv_flags & INVERT_LED_PWR)
      n |= 0x10;
    if (private_data.drv_flags & INVERT_MII_PWR)
      n |= 0x4000;
    outw(n, ioaddr + Wn2_ResetOptions);
  }
  
  if (private_data.if_port == XCVR_10base2)
    /* Start the thinnet transceiver. We should really wait 50ms...*/
    outw(StartCoax, ioaddr + EL3_CMD);
  if (private_data.if_port != XCVR_NWAY) {
    EL3WINDOW(4);
    outw((inw(ioaddr + Wn4_Media) & ~(Media_10TP|Media_SQE)) |
	 media_tbl[private_data.if_port].media_bits, ioaddr + Wn4_Media);
  }
  
  /* RTOS and statistic??? Let's disable stats. */
  outw(StatsDisable, ioaddr + EL3_CMD);
  EL3WINDOW(6);
  for (i = 0; i < 10; i++)
    inb(ioaddr + i);
  inw(ioaddr + 10);
  inw(ioaddr + 12);
  /* New: On the Vortex we must also clear the BadSSD counter. */
  EL3WINDOW(4);
  inb(ioaddr + 12);
  
  /* Switch to register set 7 for normal use. */
  EL3WINDOW(7);
  
  if (private_data.full_bus_master_rx) { /* Boomerang bus master. */
    private_data.cur_rx = private_data.dirty_rx = private_data.cur_tx = 0;

    outl(private_data.rx_ring_dma, ioaddr + UpListPtr);
  }
  if (private_data.full_bus_master_tx) { 		/* Boomerang bus master Tx. */
    private_data.cur_tx = private_data.dirty_tx = 0;
    if (private_data.drv_flags & IS_BOOMERANG)

      outb(PKT_BUF_SZ>>8, ioaddr + TxFreeThreshold); /* Room for a packet. */
    /* Clear the Rx, Tx rings. */
    for (i = 0; i < RX_RING_SIZE; i++)	/* AKPM: this is done in vortex_open, too */
      private_data.rx_ring[i].status = 0;
    outl(0, ioaddr + DownListPtr);
  }
  /* Set receiver mode: presumably accept b-case and phys addr only. */
  rtl_3COM905C_set_rx_mode();
  outw(StatsDisable, ioaddr + EL3_CMD);   /* Turn off statistics ASAP. */
  
  outw(RxEnable, ioaddr + EL3_CMD); /* Enable the receiver. */
  outw(TxEnable, ioaddr + EL3_CMD); /* Enable transmitter. */

  /* Allow status bits to be seen. */
  private_data.status_enable = SetStatusEnb | 
    (private_data.full_bus_master_rx ? UpComplete : RxComplete) | DownComplete;
  private_data.intr_enable = SetIntrEnb | UpComplete | DownComplete;
  outw(private_data.status_enable, ioaddr + EL3_CMD);
  /* Ack all pending events, and set active indicator mask. */
  outw(AckIntr | IntLatch | TxAvailable | RxEarly | IntReq,
       ioaddr + EL3_CMD);
  outw(private_data.intr_enable, ioaddr + EL3_CMD);
  if (private_data.cb_fn_base)	
    writel(0x8000, private_data.cb_fn_base + 4);
}

/***************************************************************************************/
/* This function is used to issue commands to the card and wait till the command has   */
/* been successfuly carried out.                                                       */
/***************************************************************************************/
static void
rtl_3COM905C_issue_and_wait(int cmd)
{
  int i;

  outw(cmd, private_data.ioaddr + EL3_CMD);
  for (i = 0; i < 2000; i++) {
    if (!(inw(private_data.ioaddr + EL3_STATUS) & CmdInProgress))
      return;
  }
  
  /* OK, that didn't work.  Do it the slow way.  One second */
  for (i = 0; i < 100000; i++) {
    if (!(inw(private_data.ioaddr + EL3_STATUS) & CmdInProgress)) {
      return;
    }
    usleep(10);
  }
  printk(KERN_ERR ": command 0x%04x did not complete! Status=0x%x\n",
	  cmd, inw(private_data.ioaddr + EL3_STATUS));
}

/***************************************************************************************/
/* This functions sets the receive mode of the card.                                   */
/* Pre-Cyclone chips have no documented multicast filter, so the only                  */
/* multicast setting is to receive all multicast frames.  At least                     */
/* the chip has a very clean way to set the mode, unlike many others.                  */
/***************************************************************************************/
static void rtl_3COM905C_set_rx_mode(void)
{
  long ioaddr = private_data.ioaddr;
  int new_mode;
  
  new_mode = SetRxFilter|RxStation|RxBroadcast;//RxMulticast|RxBroadcast;
		
  outw(new_mode, ioaddr + EL3_CMD);
}

/***************************************************************************************/
/* This function is used to release the card. It is called from the close call.        */
/***************************************************************************************/
static void vortex_down(struct pci_dev *dev)
{
  long ioaddr = private_data.ioaddr;
  
  /* Turn off statistics ASAP.  We update private_data->stats below. */
  outw(StatsDisable, ioaddr + EL3_CMD);
  
  /* Disable the receiver and transmitter. */
  outw(RxDisable, ioaddr + EL3_CMD);
  outw(TxDisable, ioaddr + EL3_CMD);
  
  if (private_data.if_port == XCVR_10base2)
    /* Turn off thinnet power.  Green! */
    outw(StopCoax, ioaddr + EL3_CMD);
  
  outw(SetIntrEnb | 0x0000, ioaddr + EL3_CMD);
  
  if (private_data.full_bus_master_rx)
    outl(0, ioaddr + UpListPtr);
  if (private_data.full_bus_master_tx)
    outl(0, ioaddr + DownListPtr);
  
  if (private_data.pdev && private_data.enable_wol) {
    rt_pci_save_state(private_data.pdev, private_data.power_state);
    rtl_3COM905C_acpi_set_WOL();
  }
}

/***************************************************************************************/
/* This function is used to release the card.                                          */
/***************************************************************************************/
static int eth_close(struct pci_dev *dev)
{
  int i;
  
  vortex_down(dev);

  if(private_data.must_free_irq){
    rtl_free_irq(dev->irq);
    private_data.must_free_irq = 0;
  }
  
  if (private_data.full_bus_master_rx) { /* Free Boomerang bus master Rx buffers. */
    for (i = 0; i < RX_RING_SIZE; i++)
      if (private_data.rx_skbuff[i]) {
	rtl_free(private_data.rx_skbuff[i]);
      }
  }
  if (private_data.full_bus_master_tx) { /* Free Boomerang bus master Tx buffers. */
    for (i = 0; i < TX_RING_SIZE; i++)
      if (private_data.tx_skbuff[i]) {
	rtl_free(private_data.tx_skbuff[i]);
      }
  }

  return 0;
}

/***************************************************************************************/
/* This function is never called, but it is intended to solve errors in the card.      */
/* Handle uncommon interrupt sources.  This is a separate routine to minimize          */ 
/* the cache impact.                                                                   */
/***************************************************************************************/ 
static void vortex_error(struct pci_dev *dev, int status)
{
  long ioaddr = private_data.ioaddr;
  int do_tx_reset = 0, reset_mask = 0;
  
  if (status & RxEarly) {				/* Rx early is unused. */
    outw(AckIntr | RxEarly, ioaddr + EL3_CMD);
  }
  if (status & StatsFull) {			/* Empty statistics. */
    static int DoneDidThat;

    printk(KERN_DEBUG "%s: Updating stats.\n", dev->name);
    /* HACK: Disable statistics as an interrupt source. */
    /* This occurs when we have the wrong media type! */
    if (DoneDidThat == 0  &&
	inw(ioaddr + EL3_STATUS) & StatsFull) {
      printk(KERN_WARNING "%s: Updating statistics failed, disabling "
	     "stats as an interrupt source.\n", dev->name);
      EL3WINDOW(5);
      outw(SetIntrEnb | (inw(ioaddr + 10) & ~StatsFull), ioaddr + EL3_CMD);
      private_data.intr_enable &= ~StatsFull;
      EL3WINDOW(7);
      DoneDidThat++;
    }
  }
  if (status & IntReq) {		/* Restore all interrupt sources.  */
    outw(private_data.status_enable, ioaddr + EL3_CMD);
    outw(private_data.intr_enable, ioaddr + EL3_CMD);
  }
  if (status & HostError) {
    u16 fifo_diag;
    EL3WINDOW(4);
    fifo_diag = inw(ioaddr + Wn4_FIFODiag);
    printk(KERN_ERR "%s: Host error, FIFO diagnostic register %4.4x.\n",
	   dev->name, fifo_diag);
    /* Adapter failure requires Tx/Rx reset and reinit. */
    if (private_data.full_bus_master_tx) {
      int bus_status = inl(ioaddr + PktStatus);
      /* 0x80000000 PCI master abort. */
      /* 0x40000000 PCI target abort. */
      
      printk(KERN_ERR "%s: PCI bus error, bus status %8.8x\n", dev->name, bus_status);
      
      /* In this case, blow the card away */
      vortex_down(dev);
      rtl_3COM905C_issue_and_wait(TotalReset | 0xff);
      vortex_up(dev);		/* AKPM: bug.  vortex_up() assumes that the rx ring is full. It may not be. */
    } else if (fifo_diag & 0x0400)
      do_tx_reset = 1;
    if (fifo_diag & 0x3000) {
      /* Reset Rx fifo and upload logic */
      rtl_3COM905C_issue_and_wait(RxReset|0x07);
      /* Set the Rx filter to the current state. */
      rtl_3COM905C_set_rx_mode();
      outw(RxEnable, ioaddr + EL3_CMD); /* Re-enable the receiver. */
      outw(AckIntr | HostError, ioaddr + EL3_CMD);
    }
  }
  
  if (do_tx_reset) {
    rtl_3COM905C_issue_and_wait(TxReset|reset_mask);
    outw(TxEnable, ioaddr + EL3_CMD);
  }
}

/***************************************************************************************/
/* This function is called during the release phase. It is mainly used to update pci   */
/* stuff.                                                                              */
/***************************************************************************************/
static void eth_remove_one (struct pci_dev *pdev)
{
  
  outw(TotalReset|0x14, private_data.ioaddr + EL3_CMD);
  
  if (private_data.pdev && private_data.enable_wol) {
    rt_pci_set_power_state(private_data.pdev, 0);	/* Go active */
    if (private_data.pm_state_valid)
      pci_restore_state(private_data.pdev, private_data.power_state);
  }
  
  pci_free_consistent(pdev,
		      sizeof(struct boom_rx_desc) * RX_RING_SIZE,
		      private_data.rx_ring,
		      private_data.rx_ring_dma);

  pci_free_consistent(pdev,
		      sizeof(struct boom_tx_desc) * TX_RING_SIZE,
		      private_data.tx_ring,
		      private_data.tx_ring_dma);

  if (private_data.must_free_region){
    release_region(private_data.ioaddr, private_data.io_size);
    private_data.must_free_region = 0;
  }
}

/***************************************************************************************/
/* This function is used to send a packet. It writes into the internal buffers of the  */
/* card.                                                                               */
/***************************************************************************************/
static int rt_eth_send_packet(const char *buffer, size_t size)
{
  long ioaddr = private_data.ioaddr;
  unsigned char *buff;
  int entry = private_data.cur_tx % TX_RING_SIZE;
  int previous = (private_data.cur_tx + TX_RING_SIZE - 1) % TX_RING_SIZE;
  struct boom_tx_desc *actual = &private_data.tx_ring[entry];

  rt_3c905c_writting = 1;

  /* Wait for the stall to complete. */
  rtl_3COM905C_issue_and_wait(DownStall);
  
  buff = private_data.tx_skbuff[entry];
  
  if(buff)
    memcpy(buff, buffer, size);
  
  actual->length = cpu_to_le32(size | LAST_FRAG);
  
  if (inl(ioaddr + DownListPtr) == 0) {
    outl(private_data.tx_ring_dma + entry * sizeof(struct boom_tx_desc), ioaddr + DownListPtr);
    {
      int tmp = previous;

      while((private_data.tx_ring[tmp].next != 0) && (tmp != entry)){
	private_data.tx_ring[tmp].next = 0;
	tmp = (tmp + TX_RING_SIZE - 1) % TX_RING_SIZE;
      }
    }
    private_data.queued_packet++;
  }else
    private_data.tx_ring[previous].next = cpu_to_le32(private_data.tx_ring_dma + sizeof(struct boom_tx_desc) * entry);      
  
  outw(DownUnstall, ioaddr + EL3_CMD);
  
  private_data.cur_tx++;
  
  rt_3c905c_writting = 0;

  return size;
}

/* MII transceiver control section.
   Read and write the MII registers using software-generated serial
   MDIO protocol.  See the MII specifications or DP83840A data sheet
   for details. */

/* The maximum data clock rate is 2.5 Mhz.  The minimum timing is usually
   met by back-to-back PCI I/O cycles, but we insert a delay to avoid
   "overclocking" issues. */
#define rtl_3COM905C_mdio_delay() inl(mdio_addr)

#define MDIO_SHIFT_CLK	0x01
#define MDIO_DIR_WRITE	0x04
#define MDIO_DATA_WRITE0 (0x00 | MDIO_DIR_WRITE)
#define MDIO_DATA_WRITE1 (0x02 | MDIO_DIR_WRITE)
#define MDIO_DATA_READ	0x02
#define MDIO_ENB_IN		0x00

/***************************************************************************************/
/* Generate the preamble required for initial synchronization and a few older          */
/* transceivers.                                                                       */
/***************************************************************************************/
static void rtl_3COM905C_mdio_sync(long ioaddr, int bits)
{
	long mdio_addr = ioaddr + Wn4_PhysicalMgmt;

	/* Establish sync by sending at least 32 logic ones. */
	while (-- bits >= 0) {
		outw(MDIO_DATA_WRITE1, mdio_addr);
		rtl_3COM905C_mdio_delay();
		outw(MDIO_DATA_WRITE1 | MDIO_SHIFT_CLK, mdio_addr);
		rtl_3COM905C_mdio_delay();
	}
}

/***************************************************************************************/
static int rtl_3COM905C_mdio_read(int phy_id, int location)
{
	int i;
	long ioaddr = private_data.ioaddr;
	int read_cmd = (0xf6 << 10) | (phy_id << 5) | location;
	unsigned int retval = 0;
	long mdio_addr = ioaddr + Wn4_PhysicalMgmt;

	if (mii_preamble_required)
		rtl_3COM905C_mdio_sync(ioaddr, 32);

	/* Shift the read command bits out. */
	for (i = 14; i >= 0; i--) {
		int dataval = (read_cmd&(1<<i)) ? MDIO_DATA_WRITE1 : MDIO_DATA_WRITE0;
		outw(dataval, mdio_addr);
		rtl_3COM905C_mdio_delay();
		outw(dataval | MDIO_SHIFT_CLK, mdio_addr);
		rtl_3COM905C_mdio_delay();
	}
	/* Read the two transition, 16 data, and wire-idle bits. */
	for (i = 19; i > 0; i--) {
		outw(MDIO_ENB_IN, mdio_addr);
		rtl_3COM905C_mdio_delay();
		retval = (retval << 1) | ((inw(mdio_addr) & MDIO_DATA_READ) ? 1 : 0);
		outw(MDIO_ENB_IN | MDIO_SHIFT_CLK, mdio_addr);
		rtl_3COM905C_mdio_delay();
	}

	return retval & 0x20000 ? 0xffff : retval>>1 & 0xffff;
}

/***************************************************************************************/
static void rtl_3COM905C_mdio_write(int phy_id, int location, int value)
{
	long ioaddr = private_data.ioaddr;
	int write_cmd = 0x50020000 | (phy_id << 23) | (location << 18) | value;
	long mdio_addr = ioaddr + Wn4_PhysicalMgmt;
	int i;

	if (mii_preamble_required)
		rtl_3COM905C_mdio_sync(ioaddr, 32);

	/* Shift the command bits out. */
	for (i = 31; i >= 0; i--) {
		int dataval = (write_cmd&(1<<i)) ? MDIO_DATA_WRITE1 : MDIO_DATA_WRITE0;
		outw(dataval, mdio_addr);
		rtl_3COM905C_mdio_delay();
		outw(dataval | MDIO_SHIFT_CLK, mdio_addr);
		rtl_3COM905C_mdio_delay();
	}
	/* Leave the interface idle. */
	for (i = 1; i >= 0; i--) {
		outw(MDIO_ENB_IN, mdio_addr);
		rtl_3COM905C_mdio_delay();
		outw(MDIO_ENB_IN | MDIO_SHIFT_CLK, mdio_addr);
		rtl_3COM905C_mdio_delay();
	}
	return;
}


