/*
 * Copywrite 2004 Der Herr Hofrat
 * License GPL V2
 * Author der.herr@hofr.at
 */
/*
 * scheduling jitter measure based on rt_process.c modified to write a complete
 * historgram to a file on module exit. The main advantage of this is that it
 * allows a fine grain jitter resolution that can be processed by statistic
 * tools which was not posible in rt_process.c due to the blocks of n samples
 * being reported (atleast I did not figure out how to calculate the standard
 * deviation of M samples if I have the standard deviation of each block of
 * M/N samples only (excluding the trivial case of M==N)).
 *
 * Compile time parameters to set should be limited to
 * GRAPH_SIZE the default is 1000 if set substantially larger and actually
 *   used then there are side effects on the memory subsystem that will
 *   distort results a bit
 * GRAPH_SHIFT defines how fine grain the graph is - a shift value that is 
 * larger than 8 probably makes little sense as the timer resolution is not
 * that high - the default 6 gives a resonable window from 0 to 64 
 * microseconds with fair resolution.
 */ 
#include <rtl.h>
#include <time.h>
#include <rtl_sched.h>
#include <rtl_sync.h>
#include <pthread.h>
#include <unistd.h>
#include <rtl_debug.h>
#include <rtl_cache.h>
#include <errno.h>

#include <linux/module.h>
#include <linux/slab.h>
#include <asm/uaccess.h>

int ntests=500;
int period=1000000;
int bperiod=3100000;
int mode=0;
int absolute=0;
int advance=0;
char *file = NULL;

MODULE_PARM(period,"i");
MODULE_PARM(bperiod,"i");
MODULE_PARM(ntests,"i");
MODULE_PARM(mode,"i");
MODULE_PARM(absolute,"i");
MODULE_PARM(advance,"i");
MODULE_PARM(file,"s");

MODULE_AUTHOR("Der Herr Hofrat <der.herr@hofr.at>");
MODULE_LICENSE("GPL v2");
MODULE_DESCRIPTION("scheduling jitter histogram");

#define LOOPS 10000
#define GRAPHSIZE 1000
#define GRAPH_SHIFT 6 /* the granularity of the histogram */
unsigned long *graph;
int out_of_bounds=0;
pthread_t thread;

static int orgfsuid,orgfsgid;
static mm_segment_t orgfs;

void 
kspace_init(void)
{
  orgfsuid=current->fsuid;
  orgfsgid=current->fsgid;
  current->fsuid=current->fsgid=0;

  orgfs=get_fs();
  set_fs(KERNEL_DS);
}

void
kspace_releas(void)
{
  set_fs(orgfs);
  current->fsuid=orgfsuid;
  current->fsgid=orgfsgid;
}


int kcat(char *dst_file,char *data,int size)
{
  struct file *fd0;
  int retval=0;

  kspace_init();
  if(dst_file&&*dst_file){
    fd0 = filp_open(dst_file,O_WRONLY|O_TRUNC|O_CREAT,0644);
    if (IS_ERR(fd0)){
       printk("kcat: Error %ld opening %s\n",-PTR_ERR(fd0),dst_file);
       return 0;
    }else{
      if(fd0->f_op&&fd0->f_op->write){
        int index=0;
        while((index<size)&&((retval=fd0->f_op->write(fd0,data+index,size-index,&fd0->f_pos))>0)){
          index+=retval;
        }
        if(index<size){
          printk("kcat: Write error %d\n",-retval);
        }
      }else{
        printk("kcp: %s does not have a write method\n",dst_file);
      }
    }
    retval=filp_close(fd0,NULL);
    if(retval){
       printk("kcat: Error %d closing %s\n",-retval,dst_file);
    }
  } else {
    printk("kcat: No valid destination file\n");
    return -1;
  }
  kspace_releas();
  return 0;
}

void *thread_code(void *param)
{
	hrtime_t diff;
	hrtime_t expected; 
	hrtime_t now;
	hrtime_t last_time = 0;
	hrtime_t min_diff;
	hrtime_t max_diff;
	int i,loops;
	int cnt = 0;
	int cpu_id = rtl_getcpuid();

	cache_setup();

	rtl_printf ("Measurement task starts on CPU %d\n", cpu_id);
	if (mode) {
		int ret = rtl_setclockmode (CLOCK_REALTIME, RTL_CLOCK_MODE_PERIODIC, period);

		if (ret != 0) {
			conpr("Setting periodic mode failed\n");
			mode = 0;
		}
	} else {

		rtl_setclockmode (CLOCK_REALTIME, RTL_CLOCK_MODE_ONESHOT, 0);
	}

	expected = clock_gethrtime(CLOCK_REALTIME) + 2 * (hrtime_t) period;


	if (advance) {
		rtl_stop_interrupts(); /* Be careful with this! The task won't be preempted by anything else. This is probably only appropriate for small high-priority tasks. */
	}

	/* first cycle */
	clock_nanosleep (CLOCK_REALTIME, TIMER_ABSTIME, hrt2ts(expected - advance), NULL);
	expected += period;
	now = clock_gethrtime(CLOCK_MONOTONIC);
	last_time = now;

	for(loops=1;loops<LOOPS;loops++){
		min_diff = 2000000000;
		max_diff = -2000000000;

		for (i = 0; i < ntests; i++) {
			++cnt;
			clock_nanosleep (CLOCK_REALTIME, TIMER_ABSTIME, hrt2ts(expected - advance), NULL);

			now = clock_gethrtime(CLOCK_MONOTONIC);
			if (absolute && advance && !mode) {
				if (now < expected) {
					rtl_delay (expected - now);
				}
				now = clock_gethrtime(CLOCK_MONOTONIC);
			}
			if (absolute) {
				diff = now - expected;
			} else {
				diff = now - last_time - period;
				if (diff < 0) {
					diff = -diff;
				}
			}
			if(diff>>GRAPH_SHIFT < GRAPHSIZE){
				graph[diff>>GRAPH_SHIFT]++;
			} else {
				out_of_bounds=1;
			}
			expected += period;
			last_time = now;
		}
	};

	rtl_printf("done");
	return 0;
}

pthread_t background_threadid;

void *background_thread(void *param)
{
	hrtime_t next = clock_gethrtime(CLOCK_REALTIME);
	while (1) {
		hrtime_t t = gethrtime ();
		next += bperiod;
		/* the measurement task should preempt the following loop */
//		__asm__ __volatile__("invlpg %0": :"m" (*(char *) &next));
		while (gethrtime() < t + bperiod * 2 / 3){
			;
		}
		clock_nanosleep (CLOCK_REALTIME, TIMER_ABSTIME, hrt2ts(next), NULL);
	}
}


int mtrr_reg;
unsigned long phys_add;

int init_module(void)
{
	pthread_attr_t attr;
	struct sched_param sched_param;
	int thread_status;

	graph=kmalloc(GRAPHSIZE*sizeof(unsigned long),GFP_KERNEL);
	memset(graph,0,GRAPHSIZE*sizeof(unsigned long));

	printk("RTLinux measurement module on CPU %d\n",rtl_getcpuid());
	pthread_attr_init (&attr);
	if (rtl_cpu_exists(1)) {
		pthread_attr_setcpu_np(&attr, 1);
	}
	sched_param.sched_priority = 1;
	pthread_attr_setschedparam (&attr, &sched_param);
	printk("About to thread create\n");
	thread_status = pthread_create (&thread,  &attr, thread_code, (void *)1);
	if (thread_status != 0) {
		printk("failed to create RT-thread: %d\n", thread_status);
		return -1;
	} else {
		printk("created RT-thread\n");
	}

	if (bperiod) {
		pthread_create (&background_threadid,  NULL, background_thread, NULL);
	}
	return 0;
}

void cleanup_module(void)
{
	char *data;
	int size,len,i;

	/* allocate buffer for human redable data */
	size=GRAPHSIZE*sizeof(unsigned long)*4;
	data=kmalloc(size,GFP_KERNEL);
	memset(data,0,size);

	/* shut down rt-thread */
	pthread_cancel (thread);
	pthread_join (thread, NULL);
	if (bperiod) {
		pthread_cancel (background_threadid);
		pthread_join (background_threadid, NULL);
	}

	/* build data file */
	len=0;
	if(!out_of_bounds){
		for(i=0;i<GRAPHSIZE;i++){
			len+=sprintf(data+len,"%6d %10ld\n",i<<GRAPH_SHIFT,graph[i]);
		}
	} else {
		len=sprintf(data,"max jitter out of bounds (period=%d,bperiod=%d,ntests=%d)\n",period,bperiod,ntests);
	}	

	/* dump to file */
	kcat(file,data,len);

	kfree(graph);
}

