/*
 * Copyright (c) 2001, Swedish Institute of Computer Science.
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution. 
 * 3. Neither the name of the Institute nor the names of its contributors 
 *    may be used to endorse or promote products derived from this software 
 *    without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE 
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE. 
 *
 */

/*********************************************************************************/
/* This file is based in one file part of the lwIP TCP/IP stack. The file is:    */
/* ethernetif.c                                                                  */
/* which author is: Adam Dunkels <adam@sics.se>                                  */
/*                                                                               */
/* This file has been modified by Sergio Perez Alcaiz <serpeal@upvnet.upv.es>   */
/*            Departamento de Informtica de Sistemas y Computadores             */
/*            Universidad Politcnica de Valencia                                */
/*            Valencia (Spain)                                                   */
/*                                                                               */
/* The RTL-lwIP project has been supported by the Spanish Government Research    */
/* Office (CICYT) under grant TIC2002-04123-C03-03                               */
/*                                                                               */
/* Copyright (c) March, 2003 SISTEMAS DE TIEMPO REAL EMPOTRADOS, FIABLES Y       */
/* DISTRIBUIDOS BASADOS EN COMPONENTES                                           */
/*                                                                               */
/*  This program is free software; you can redistribute it and/or modify         */
/*  it under the terms of the GNU General Public License as published by         */
/*  the Free Software Foundation; either version 2 of the License, or            */
/*  (at your option) any later version.                                          */
/*                                                                               */
/*  This program is distributed in the hope that it will be useful,              */
/*  but WITHOUT ANY WARRANTY; without even the implied warrabnty of              */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                */
/*  GNU General Public License for more details.                                 */
/*                                                                               */
/*  You should have received a copy of the GNU General Public License            */
/*  along with this program; if not, write to the Free Software                  */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA    */
/*                                                                               */
/*  Linking RTL-lwIP statically or dynamically with other modules is making a    */
/*  combined work based on RTL-lwIP.  Thus, the terms and conditions of the GNU  */
/*  General Public License cover the whole combination.                          */
/*                                                                               */
/*  As a special exception, the copyright holders of RTL-lwIP give you           */
/*  permission to link RTL-lwIP with independent modules that communicate with   */
/*  RTL-lwIP solely through the interfaces, regardless of the license terms of   */
/*  these independent modules, and to copy and distribute the resulting combined */
/*  work under terms of your choice, provided that every copy of the combined    */
/*  work is accompanied by a complete copy of the source code of RTL-lwIP (the   */
/*  version of RTL-lwIP used to produce the combined work), being distributed    */
/*  under the terms of the GNU General Public License plus this exception.  An   */
/*  independent module is a module which is not derived from or based on         */
/*  RTL-lwIP.                                                                    */
/*                                                                               */
/*  Note that people who make modified versions of RTL-lwIP are not obligated to */
/*  grant this special exception for their modified versions; it is their choice */
/*  whether to do so.  The GNU General Public License gives permission to        */
/*  release a modified version without this exception; this exception also makes */
/*  it possible to release a modified version which carries forward this         */
/*  exception.                                                                   */
/*********************************************************************************/

#include "lwip/opt.h"
#include "lwip/def.h"
#include "lwip/ip.h"
#include "lwip/mem.h"
#include "lwip/pbuf.h"
#include "lwip/sys.h"
#include "bcopy.h"
#include "netif/arp.h"
#include <time.h>
#include <rtl_sched.h>
#include <rtl_fifo.h>

#ifdef __RTFIFOOPTS__

#define DEVWRITE "/dev/rtf0"
#define DEVREAD "/dev/rtf1"

#define IFNAME0 'r'
#define IFNAME1 't'

#define SETBYTEONE(a,b) (a=(b & 0x000F))
#define SETBYTETWO(a,b) (a=(b & 0x00F0)) 
#define SETBYTETHREE(a,b) (a=(b & 0x0F00))
#define SETBYTEFOUR(a,b) (a=(b & 0xF000))

static const struct eth_addr ethbroadcast = {{0xff,0xff,0xff,0xff,0xff,0xff}};

struct rtfifoif {
  struct eth_addr *ethaddr;
  int writefd;
  int readfd;
};

/* Forward declarations. */
static void  rtfifoif_input(struct netif *netif);
static err_t rtfifoif_output(struct netif *netif, struct pbuf *p, struct ip_addr *ipaddr);
static void rtfifoif_thread(void *arg);

/*-----------------------------------------------------------------------------------*/
static void
low_level_init(struct netif *netif)
{
  struct rtfifoif *rtfifoif;
  rtfifoif = netif->state;
  
  /* Obtain MAC address from network interface. */
  string2mac(rtfifoif->ethaddr,RTFIFO_MAC);

  /* We use RTfifo 0 for writing */
  rtfifoif->writefd = 0;
  /* We use RTfifo 1 for reading */
  rtfifoif->readfd = 1;

  sys_thread_new(rtfifoif_thread, netif, RTFIFO_THREAD_PERIOD);

}

/*-----------------------------------------------------------------------------------*/
static void 
rtfifoif_thread(void *arg){
  struct netif *netif;
  struct rtfifoif *rtfifoif;
 
  netif = arg;
  rtfifoif = netif->state;
  
  while(1) {
    pthread_wait_np();
    /* Wait for a packet to arrive. */
    while(rtf_isempty(rtfifoif->readfd)){
      usleep(RTFIFO_DELAY);
    }
	  
    /* Handle incoming packet. */
    rtfifoif_input(netif);
  }
}

/*-----------------------------------------------------------------------------------*/
/*
 * low_level_output():
 *
 * Should do the actual transmission of the packet. The packet is
 * contained in the pbuf that is passed to the function. This pbuf
 * might be chained.
 *
 */
/*-----------------------------------------------------------------------------------*/
static err_t
low_level_output(struct rtfifoif *rtfifoif, struct pbuf *p)
{
  struct pbuf *q;
  char buf[1500];
  char *bufptr;
  unsigned char size_of_packet[4];
  
  /* initiate transfer(); */
  
  bufptr = &buf[0];
  
  for(q = p; q != NULL; q = q->next) {
    /* Send the data from the pbuf to the interface, one pbuf at a
       time. The size of the data in each pbuf is kept in the ->len
       variable. */    
    /* send data from(q->payload, q->len); */
    bcopy(q->payload, bufptr, q->len);
    bufptr += q->len;
  }
  
  SETBYTEONE(size_of_packet[0],p->tot_len);
  SETBYTETWO(size_of_packet[1],p->tot_len); 
  SETBYTETHREE(size_of_packet[2],p->tot_len);
  SETBYTEFOUR(size_of_packet[3],p->tot_len);

  if(rtf_put(rtfifoif->writefd, &size_of_packet, 4) == -1) {
    rtl_printf("low_level_output: rtf_put error");
  }else{
    if(rtf_put(rtfifoif->writefd, buf, p->tot_len) == -1) {
      rtl_printf("low_level_output: rtf_put error");
    }
  }
  
  return ERR_OK;
}
/*-----------------------------------------------------------------------------------*/
/*
 * low_level_input():
 *
 * Should allocate a pbuf and transfer the bytes of the incoming
 * packet from the interface into the pbuf.
 *
 */
/*-----------------------------------------------------------------------------------*/
static struct pbuf *
low_level_input(struct rtfifoif *rtfifoif)
{
  struct pbuf *p, *q;
  u16_t len;
  char buf[1500];
  char *bufptr;
  unsigned char size_of_packet[4];
  int packet_size;

  /* Obtain the size of the packet and put it into the "len"
     variable. */

  rtf_get(rtfifoif->readfd, &size_of_packet, 4);
  
  packet_size = size_of_packet[0] + size_of_packet[1] + size_of_packet[2] + size_of_packet[3];

  /* Wait for a packet to arrive. */
  while(rtf_isempty(rtfifoif->readfd)){
    usleep(10);
  }

  len = rtf_get(rtfifoif->readfd, buf,(int) packet_size);


  /* We allocate a pbuf chain of pbufs from the pool. */
  p = pbuf_alloc(PBUF_LINK, len, PBUF_POOL);
  
  if(p != NULL) {
    /* We iterate over the pbuf chain until we have read the entire
       packet into the pbuf. */
    bufptr = &buf[0];
    for(q = p; q != NULL; q = q->next) {
      /* Read enough bytes to fill this pbuf in the chain. The
         avaliable data in the pbuf is given by the q->len
         variable. */
      /* read data into(q->payload, q->len); */
      bcopy(bufptr, q->payload, q->len);
      bufptr += q->len;
    }
    /* acknowledge that packet has been read(); */
  } else {
    rtl_printf("rtfifoif: Can't allocate memory for pbuf\n");
  }

  return p;  
}

/*-----------------------------------------------------------------------------------*/
/*
 * fifoif_output():
 *
 * This function is called by the TCP/IP stack when an IP packet
 * should be sent. It calls the function called low_level_output() to
 * do the actuall transmission of the packet.
 *
 */
/*-----------------------------------------------------------------------------------*/
static err_t
rtfifoif_output(struct netif *netif, struct pbuf *p,
		  struct ip_addr *ipaddr)
{
  struct rtfifoif *rtfifoif;
  struct pbuf *q;
  struct eth_hdr *ethhdr;
  struct eth_addr *dest, mcastaddr;
  struct ip_addr *queryaddr;
  err_t err;
  u8_t i;
  
  rtfifoif = netif->state;

  /* Make room for Ethernet header. */
  if(pbuf_header(p, sizeof(struct eth_hdr)) != 0) {
    /* The pbuf_header() call shouldn't fail, but we allocate an extra
       pbuf just in case. */
    q = pbuf_alloc(PBUF_LINK, sizeof(struct eth_hdr), PBUF_RAM);
    if(q == NULL) {
      return ERR_MEM;
    }
    pbuf_chain(q, p);
    p = q;
  }

  /* Construct Ethernet header. Start with looking up deciding which
     MAC address to use as a destination address. Broadcasts and
     multicasts are special, all other addresses are looked up in the
     ARP table. */
  queryaddr = ipaddr;
  if(ip_addr_isany(ipaddr) ||
     ip_addr_isbroadcast(ipaddr, &(netif->netmask))) {
    dest = (struct eth_addr *)&ethbroadcast;
  } else if(ip_addr_ismulticast(ipaddr)) {
    /* Hash IP multicast address to MAC address. */
    mcastaddr.addr[0] = 0x01;
    mcastaddr.addr[1] = 0x0;
    mcastaddr.addr[2] = 0x5e;
    mcastaddr.addr[3] = ip4_addr2(ipaddr) & 0x7f;
    mcastaddr.addr[4] = ip4_addr3(ipaddr);
    mcastaddr.addr[5] = ip4_addr4(ipaddr);
    dest = &mcastaddr;
  } else {
    if(ip_addr_maskcmp(ipaddr, &(netif->ip_addr), &(netif->netmask))) {
      /* Use destination IP address if the destination is on the same
         subnet as we are. */
      queryaddr = ipaddr;
    } else {
      /* Otherwise we use the default router as the address to send
         the Ethernet frame to. */
      queryaddr = &(netif->gw);
    }
    dest = arp_lookup(queryaddr);
  }

  /* If the arp_lookup() didn't find an address, we send out an ARP
     query for the IP address. */
  if(dest == NULL) {
    q = arp_query(netif, rtfifoif->ethaddr, queryaddr);
    if(q != NULL) {
      err = low_level_output(rtfifoif, q);
      pbuf_free(q);
      return err;
    }
    return ERR_MEM;
  }
  ethhdr = p->payload;
  
  for(i = 0; i < 6; i++) {
    ethhdr->dest.addr[i] = dest->addr[i];
    ethhdr->src.addr[i] = rtfifoif->ethaddr->addr[i];
  }
  
  ethhdr->type = htons(ETHTYPE_IP);
  
  return low_level_output(rtfifoif, p);

}
/*-----------------------------------------------------------------------------------*/
/*
 * rtfifoif_input():
 *
 * This function should be called when a packet is ready to be read
 * from the interface. It uses the function low_level_input() that
 * should handle the actual reception of bytes from the network
 * interface.
 *
 */
/*-----------------------------------------------------------------------------------*/
static void
rtfifoif_input(struct netif *netif)
{
  struct rtfifoif *rtfifoif;
  struct eth_hdr *ethhdr;
  struct pbuf *p;


  rtfifoif = netif->state;
  
  p = low_level_input(rtfifoif);
  
  if(p == NULL) {
    return;
  }
  ethhdr = p->payload;

  switch(htons(ethhdr->type)) {
  case ETHTYPE_IP:
    arp_ip_input(netif, p);
    pbuf_header(p, -14);
    if(ip_lookup(p->payload, netif)) {
      netif->input(p, netif);
    }
    break;
  case ETHTYPE_ARP:

    p = arp_arp_input(netif, rtfifoif->ethaddr, p);
    if(p != NULL) {
      low_level_output(rtfifoif, p);
      pbuf_free(p);
    }
    break;
  default:
    pbuf_free(p);
    break;
  }
}
/*-----------------------------------------------------------------------------------*/
static void
arp_timer(void *arg)
{
  arp_tmr();
  sys_timeout(ARP_TMR_INTERVAL, (sys_timeout_handler)arp_timer, NULL);
}
/*-----------------------------------------------------------------------------------*/
/*
 * rtfifoif_init():
 *
 * Should be called at the beginning of the program to set up the
 * network interface. It calls the function low_level_init() to do the
 * actual setup of the hardware.
 *
 */
/*-----------------------------------------------------------------------------------*/
void
rtfifoif_init(struct netif *netif)
{
  struct rtfifoif *rtfifoif;
    
  rtfifoif = mem_malloc(sizeof(struct rtfifoif));
  netif->state = rtfifoif;
  netif->name[0] = IFNAME0;
  netif->name[1] = IFNAME1;
  netif->output = rtfifoif_output;
  netif->linkoutput = low_level_output;
  
  rtfifoif->ethaddr = (struct eth_addr *)&(netif->hwaddr[0]);
  
  low_level_init(netif);
  arp_init();
  
  _timeout(ARP_TMR_INTERVAL, (sys_timeout_handler)arp_timer, NULL);
}
/*-----------------------------------------------------------------------------------*/
#endif //__RTFIFOOPTS__
